from .error import RouteHandleError


class Route():
    '''
    Specify a route in a map
    '''

    def __init__(self, path, method, req_type, handle):
        '''
        :param str path:
        :param str method:
        :param str req_type:
        :param function handle:
        '''

        if not callable(handle):
            raise RouteHandleError(handle)

        self.path = path
        self.method = method
        self.req_type = req_type
        self.handle = handle



import re
from clink.type import HTTP_GET, HTTP_POST, HTTP_PUT, HTTP_DELETE, HTTP_PATCH
from clink.mime import MIME_JSON
from clink.com.label import write_stamp
from .error import RoutePathError
from .type import CtlMethod

CTL_PATH_ATTR = 'ctl_path'
CTL_METHOD_ATTR = 'ctl_method'

_ROUTE_PATH_REGEX = re.compile('^(/[^/]*|/[^/]+(/[^/]+)+)$')


def path(path):
    '''
    Specify a path to controller

    :param str path:
    '''

    if _ROUTE_PATH_REGEX.match(path) is None:
        raise RoutePathError(path, _ROUTE_PATH_REGEX)

    def decorator_fn(target):
        write_stamp(target, CTL_PATH_ATTR, path)
        return target
    return decorator_fn


def map(path, method, req_type):
    '''
    Specify a controller's method

    :param str path:
    :param str method:
    :param str req_type:
    '''

    if _ROUTE_PATH_REGEX.match(path) is None:
        raise RoutePathError(path, _ROUTE_PATH_REGEX)

    def decorator_fn(target):
        ctl_method = CtlMethod(path, method, req_type)
        write_stamp(target, CTL_METHOD_ATTR, ctl_method)
        return target
    return decorator_fn


def get(path):
    '''
    Specify a GET controller's method

    :param str path:
    '''

    return map(path, HTTP_GET, None)


def post(path, req_type=MIME_JSON):
    '''
    Map a POST controller's method

    :param str path:
    :param str req_type:
    '''

    return map(path, HTTP_POST, req_type)


def put(path, req_type=MIME_JSON):
    '''
    Map a PUT controller's method

    :param str path:
    :param str req_type:
    '''

    return map(path, HTTP_PUT, req_type)


def patch(path, req_type=MIME_JSON):
    '''
    Map a PATCH controller's method

    :param str path:
    :param str req_type:
    '''

    return map(path, HTTP_PATCH, req_type)


def delete(path, req_type=MIME_JSON):
    '''
    Map a DELETE controller's method

    :param str path:
    '''

    return map(path, HTTP_DELETE, req_type)



from clink.com import read_stamp
from clink.iface import ILv2Handler
from clink.error.http import Http400Error, Http404Error, Http405Error

from .error import RouteExistError
from .route import Route
from .mapper import CTL_PATH_ATTR, CTL_METHOD_ATTR


class Router(ILv2Handler):
    '''
    Store and find routes
    '''

    def __init__(self, routes=[]):
        '''
        :param list<Router> routes:
        '''

        self._map = {}
        self.add_routes(routes)

    def add_ctl(self, ctl):
        '''
        Collect routing information from controller

        :param object ctl:
        '''

        ctl_type = type(ctl)
        ctl_path = read_stamp(ctl, CTL_PATH_ATTR)
        for attr_name in dir(ctl_type):
            ctl_attr = getattr(ctl_type, attr_name)
            try:
                ctl_method = read_stamp(ctl_attr, CTL_METHOD_ATTR)
                abs_path = ctl_path + ctl_method.path
                abs_path = abs_path.replace('//', '/')
                if len(abs_path) > 1 and abs_path[-1] == '/':
                    abs_path = abs_path[:len(abs_path) - 1]
                route = Route(
                    abs_path, ctl_method.method, ctl_method.req_type,
                    getattr(ctl, attr_name)
                )
                self.add_route(route)
            except KeyError:
                pass
            except TypeError:
                pass

    def add_route(self, route):
        '''
        Put route into map

        :param Route route:
        :raise RouteExistError:
        '''
        if route.path not in self._map:
            self._map[route.path] = {}
        if route.method not in self._map:
            self._map[route.path][route.method] = {}
        if route.req_type in self._map[route.path][route.method]:
            raise RouteExistError(route)
        self._map[route.path][route.method][route.req_type] = route

    def add_routes(self, routes):
        '''
        Put routes into map

        :param list[Route] routes:
        '''

        for route in routes:
            self.add_route(route)

    def handle(self, req):
        '''
        Find handle which match with request

        :param Request req:
        :rtype: function
        :raise Http400Error:
        :raise Http404Error:
        :raise Http405Error:
        '''

        path = req.path
        method = req.method

        if path not in self._map:
            raise Http404Error(req)
        if method not in self._map[path]:
            raise Http405Error(req)
        if req.content_type not in self._map[path][method]:
            raise Http400Error(req)
        return self._map[path][method][req.content_type].handle



class CtlMethod():
    '''
    Specify controller method
    '''

    def __init__(self, path, method, req_type):
        '''
        :param str path:
        :param clink.type.HttpMethod method:
        :param str req_type:
        '''

        self.path = path
        self.method = method
        self.req_type = req_type
